"""
This class can be considered as a base class so you
copy the 2 classes on this file, adapt the params
(attributes) and customize the 'animate' method with
your specific animation.

The wrapper class must be always next to the generator
class because they work toghether
"""
from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.generation.manim.enums import MANIM_RENDERER
from yta_multimedia.video.generation.manim.classes.base_manim_animation_wrapper import BaseManimAnimationWrapper, ManimAnimationType
from yta_general_utils.programming.parameter_validator import NumberValidator
from manim import *
from typing import Union


# This class and the generator must be always in
# the same file because they are extremely
# connected and they are dependant
class SimpleTextManimAnimationWrapper(BaseManimAnimationWrapper):
    """
    Simple Text animation wrapper, used only for testing.
    """
    text: str = None
    duration: float = None

    def __init__(self, text: str, duration: float):
        # TODO: This parameter validation could be done
        # by our specific parameter validator (under
        # construction yet)
        exception_messages = []

        if not text:
            exception_messages.append('No "text" parameter provided.')
        
        if not NumberValidator.is_positive_number(duration) or not NumberValidator.is_number_between(duration, 0, 100, do_include_lower_limit = False):
            exception_messages.append('The "duration" parameter provided is not a positive number between (0, 100] (zero is not valid).')

        if len(exception_messages) > 0:
            raise Exception('\n'.join([exception_message for exception_message in exception_messages]))
        
        self.text = text
        self.duration = duration
        super().__init__(SimpleTextManimAnimationGenerator, ManimAnimationType.TEXT_ALPHA)

class SimpleTextManimAnimationGenerator(BaseManimAnimation):
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, parameters: dict, output_filename: Union[str, None] = None):
        """
        Build the manim animation video and stores it
        locally as the provided 'output_filename', 
        returning the final output filename.

        This method will call the '.animate()' method
        to build the scene with the instructions that
        are written in that method.
        """
        return super().generate(
            parameters,
            renderer = MANIM_RENDERER.CAIRO,
            output_filename = output_filename
        )
    
    def animate(self):
        text = Text(self.parameters['text'], font_size = 26, font = 'Minecraftia').shift(DOWN * 0).scale(1)
        self.add(text)
        # The sound duration will be set as video duration if
        # larger than the 'duration' parameter so I need to
        # look for a way to enshort the video
        #self.add_sound(GoogleDriveResource('https://drive.google.com/file/d/1Dzeb6Qae4UdpmuA6U9d6t3MXnvhzukzg/view?usp=sharing').download(create_temp_filename('tmp.mp3')))
        self.wait(self.parameters['duration'])

    #     def animate_one(self):
    #         """
    #         This code will generate the manim animation and belongs to the
    #         Scene manim object.
    #         """
    #         text = Text(self.parameters['text'], font_size = 140, stroke_width = 2.0, font = 'Haettenschweiler').shift(DOWN * 0).scale(0.001)
    #         self.wait(1 / 60)
    #         self.add(text)
    #         self.add_sound(GoogleDriveResource('https://drive.google.com/file/d/1WPS8uWB1LTuzPzxQ2Zcp1FpwvLM3fhM5/view?usp=sharing').download(create_temp_filename('tmp.mp3')))
    #         self.play(text.animate.scale(1000), run_time = 49 / 60)
    #         self.play(Rotate(text, 0.03), run_time = 3 / 60)
    #         self.play(Rotate(text, -0.04), run_time = 3 / 60)
    #         self.play(Rotate(text, -0.02), run_time = 3 / 60)
    #         self.play(Rotate(text, 0.04), run_time = 3 / 60)
    #         self.play(Rotate(text, -0.01), run_time = 3 / 60)
    #         self.play(Rotate(text, 0.03), run_time = 3 / 60)
    #         self.play(Rotate(text, -0.04), run_time = 3 / 60)
    #         self.play(Rotate(text, -0.02), run_time = 3 / 60)
    #         self.play(Rotate(text, 0.04), run_time = 3 / 60)
    #         self.play(Rotate(text, -0.01), run_time = 3 / 60)
    #         #self.add_sound(TOUSE_ABSOLUTE_PATH + 'sounds/xp_error.mp3')
    #         #self.play(AddTextLetterByLetter(text), run_time = self.parameters['duration'])
            
    #         #simple_play_animation(self, Write, text, self.parameters['duration'])

    #     def animate_two(self):
    #         """
    #         This code will generate the manim animation and belongs to the
    #         Scene manim object.
    #         """
    #         text = Text(self.parameters['text'], font_size = 140, stroke_width = 2.0, font = 'Haettenschweiler').shift(DOWN * 0).scale(7 / 10)
    #         self.wait(1 / 60)
    #         self.add(text)
    #         self.play(text.animate.scale(10 / 7), run_time = 6 / 60)
    #         self.play(ApplyWave(text), run_time = 30 / 60)

    #     def animate(self):
    #         text = Text(self.parameters['text'], font_size = 26, font = 'Minecraftia').shift(DOWN * 0).scale(1)
    #         self.add(text)
    #         # The sound duration will be set as video duration if
    #         # larger than the 'duration' parameter so I need to
    #         # look for a way to enshort the video
    #         #self.add_sound(GoogleDriveResource('https://drive.google.com/file/d/1Dzeb6Qae4UdpmuA6U9d6t3MXnvhzukzg/view?usp=sharing').download(create_temp_filename('tmp.mp3')))
    #         self.wait(self.parameters['duration'])

    # __all__ = [
    #     'SimpleTextManimAnimationX'
    # ]















from yta_multimedia.video.parser import VideoParser
from moviepy.Clip import Clip
from typing import Union


def subclip_video(video: Clip, start_time: float, end_time: float) -> tuple[Union[Clip, None], Clip, Union[Clip, None]]:
    """
    Subclip the provided 'video' into 3 different subclips,
    according to the provided 'start_time' and 'end_time',
    and return them as a tuple of those 3 clips. First and
    third clip could be None.

    The first clip will be None when 'start_time' is 0, and 
    the third one when the 'end_time' is equal to the given
    'video' duration.
    """
    video = VideoParser.to_moviepy(video)

    left = None if (start_time == 0) else video.with_subclip(0, start_time)
    center = video.with_subclip(start_time, end_time)
    right = None if (end_time is None) else video.with_subclip(start_time = end_time)

    return left, center, right


from skimage.filters import gaussian as skimage_gaussian
from typing import Union


class FrameFilter:
    """
    Filters to be applied on moviepy video frames so we
    can modify them by using 'clip.transform' method.

    Example below:
    ``` 
    clip.transform(
        lambda get_frame, t:
        FrameFilter.blur(get_frame, t, blur_radius = blur_radius)
    )
    ```
    
    """
    @staticmethod
    def blur(get_frame, t, blur_radius: Union[int, None] = None):
        return skimage_gaussian(get_frame(t).astype(float), sigma = blur_radius)


"""
This file has been created to preserve a part of code
that was able to project images to simulate a 3D 
effect and also being able to put those images in
front of other videos with a mask.

TODO: This is experimental, as you can see, but the
results are very nice to make some preset or similar
"""
from yta_multimedia.video.consts import MOVIEPY_SCENE_DEFAULT_SIZE


def test_perspective(angle: float, output_filename: str):
    # Use one image to calculate the projection
    import cv2
    import numpy as np

    image_to_inject = cv2.imread('C:/Users/dania/Desktop/004_youtube_video.png')
    image = cv2.imread('C:/Users/dania/Desktop/white_1920x1080.png')

    # Adapt the original image
    original_image_height, original_image_width = image_to_inject.shape[:2]
    shape_to_inject = np.array([
        [0, 0],
        [original_image_width - 1, 0],
        [original_image_width - 1, original_image_height - 1],
        [0, original_image_height - 1]
    ], dtype = 'float32')
    
    # We need to apply a resize first to make sure it fits our
    # expected dimensions, then we will adjust it and apply the
    # rotation that will change the actual dimensions as the
    # width and height changes according to the rotation, so we
    # would need to move the element again

    # First I will use static corners just to reduce its size
    # TODO: This can be dynamic according to what we need, or
    # it should be the maximum possible to fit the 1920x1080
    # dimensions without being out of it. By now it is a 3rd
    # of the region size
    desired_width = MOVIEPY_SCENE_DEFAULT_SIZE[0] / 3
    desired_height = MOVIEPY_SCENE_DEFAULT_SIZE[1] / 3
    x_start = MOVIEPY_SCENE_DEFAULT_SIZE[0] / 2 - (desired_width / 2)
    y_start = MOVIEPY_SCENE_DEFAULT_SIZE[1] / 2 - (desired_height / 2)

    # Build corners to make it possible
    corners = np.array([
        (x_start, y_start),
        (x_start + desired_width, y_start),
        (x_start + desired_width, y_start + desired_height),
        (x_start, y_start + desired_height)
    ], dtype = 'float32')

    # We need the angle in radians
    angle = np.radians(angle)
    # Rz = np.array([
    #     [np.cos(angle), -np.sin(angle), 0],
    #     [np.sin(angle), np.cos(angle), 0],
    #     [0, 0, 1]
    # ], dtype = 'float32')
    def x_rotation(angle: float):
        return np.array([
            [1, 0, 0],
            [0, np.cos(angle), -np.sin(angle)],
            [0, np.sin(angle), np.cos(angle)]
        ], dtype = 'float32')
    
    def y_rotation(angle: float):
        return np.array([
            [np.cos(angle), 0, np.sin(angle)],
            [0, 1, 0],
            [-np.sin(angle), 0, np.cos(angle)]
        ], dtype = 'float32')
    
    def z_rotation(angle: float):
        return np.array([
            [np.cos(angle), -np.sin(angle), 0],
            [np.sin(angle), np.cos(angle), 0],
            [0, 0, 1]
        ], dtype = 'float32')
    # Here we are applying the rotation we want, just play
    # with this. This will change again the image shape
    # so we need to adjust the position later
    # TODO: Play with this please
    Rz = z_rotation(angle)
    # Rz = np.array([
    #     [np.cos(angle), 0, -np.sin(angle)],
    #     [np.sin(angle), np.cos(angle), 0],
    #     [0, 0, 1]
    # ], dtype = 'float32')

    # 2D to 3D adding a 0 z coordinate
    pts3d = np.hstack((corners, np.zeros((corners.shape[0], 1))))

    # Apply rotation, so these are our new coordinates
    # with the rotation applied. We should need to make
    # a fix to put it in the middle of the scene 
    # according to the new size
    pts3d_rotadas = np.dot(pts3d, Rz.T)

    # Now we need to fix the position by size
    x_coords = [coord[0] for coord in pts3d_rotadas]
    y_coords = [coord[1] for coord in pts3d_rotadas]

    x_min = min(x_coords)
    x_max = max(x_coords)
    y_min = min(y_coords)
    y_max = max(y_coords)
    max_width_after_rotation = x_max - x_min
    max_height_after_rotation = y_max - y_min

    # This is the new x position to let it be in the
    # center
    new_start_x = MOVIEPY_SCENE_DEFAULT_SIZE[0] / 2 - max_width_after_rotation / 2
    new_start_y = MOVIEPY_SCENE_DEFAULT_SIZE[1] / 2 - max_height_after_rotation / 2
    dif_x = x_min - new_start_x
    dif_y = y_min - new_start_y

    # Add the needed amount of movement to let it be in the
    # center of the scene but keeping the rotation
    for pts3d_r in pts3d_rotadas:
        pts3d_r[0] -= dif_x
        pts3d_r[1] -= dif_y

    P = np.array([
        [1, 0, 0],  # XY projection
        [0, 1, 0],
        [0, 0, 0]
    ], dtype = 'float32')

    # Apply the projection
    pts2d_proyectadas = np.dot(pts3d_rotadas, P.T)[:, :2]
    pts2d_proyectadas = np.array(pts2d_proyectadas, dtype = 'float32')
    matrix = cv2.getPerspectiveTransform(shape_to_inject, pts2d_proyectadas)

    # Transform the perspective to fit the region
    inserted_image = cv2.warpPerspective(image_to_inject, matrix, (image.shape[1], image.shape[0]), flags = cv2.INTER_NEAREST, borderMode = cv2.BORDER_TRANSPARENT)

    # Add alpha layer to image
    # TODO: This should be not 1920x1080 but dynamic
    inserted_image = np.dstack((inserted_image, np.ones((MOVIEPY_SCENE_DEFAULT_SIZE[1], MOVIEPY_SCENE_DEFAULT_SIZE[0]), dtype = np.uint8) * 255))

    # Add transparent layer to fill poly with it
    mask = np.zeros((MOVIEPY_SCENE_DEFAULT_SIZE[1], MOVIEPY_SCENE_DEFAULT_SIZE[0]), dtype = np.uint8)

    # We fill the mask with 255 values for any pixel which
    # is inside the polygon area
    cv2.fillPoly(mask, np.int32([pts2d_proyectadas]), 255)

    # We make all the other pixels that are not the polygon
    # be transparent
    inserted_image[mask != 255, 3] = 0

    cv2.imwrite(output_filename, inserted_image)

    return inserted_image

def test():
    from moviepy import CompositeVideoClip, ImageClip, concatenate_videoclips

    clips = []
    for i in range(90):
        name = f'a_test_perspc{str(i + 1)}.png'
        test_perspective(i, name)
        clips.append(ImageClip(name, duration = 1 / 60, transparent = True).with_fps(60))

    CompositeVideoClip([
        ImageClip('C:/Users/dania/Desktop/004_youtube_video.png', duration = len(clips) * 1 / 60, transparent = True).with_fps(60),
        concatenate_videoclips(clips, is_mask = True)
    ]).write_videofile('a_test_3drotz.mp4')


from yta_general_utils.image.parser import ImageParser
from PIL import Image
from transformers import BlipProcessor, BlipForConditionalGeneration
from typing import Union


def describe_image(image: Union[str, Image.Image]):
    """
    Describes the provided 'image' by using the Salesforce blip 
    image captioning system. This method will use some
    pretrained models that are stored in 
    'C:/Users/USER/.cache/huggingface/hub', load in memory and used
    to describe the image. This process could take a couple of 
    minutes.
    """
    image = ImageParser.to_pillow(image)

    # models are stored in C:\Users\USERNAME\.cache\huggingface\hub
    processor = BlipProcessor.from_pretrained('Salesforce/blip-image-captioning-base')
    model = BlipForConditionalGeneration.from_pretrained('Salesforce/blip-image-captioning-base')
    inputs = processor(image, return_tensors = 'pt')
    out = model.generate(**inputs)
    description = processor.decode(out[0], skip_special_tokes = True)

    # TODO: Fix strange characters. I received 'a red arrow pointing up
    # to the right [SEP]' response from describing an image. What is the
    # '[SEP]' part? What does it mean? I don't want that in response.
    return description


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate import Coordinate
from yta_multimedia.video.position import Position
from yta_multimedia.video.edition.effect.moviepy.t_function import TFunctionSetPosition
from yta_multimedia.video.edition.effect.moviepy.objects import MoviepyArgument, MoviepyWith
from yta_multimedia.video.parser import VideoParser
from yta_general_utils.math.rate_functions import RateFunction
from moviepy.Clip import Clip
from typing import Union


class MoveLinearPositionEffect(Effect):
    """
    Move from A to B doing a straight line effect while moving.
    """
    # TODO: Is this working (?)
    result_must_replace = True

    def apply(cls, video: Union[Clip, str], initial_position: Union[Coordinate, Position], final_position: Union[Coordinate, Position]):
        video = VideoParser.to_moviepy(video)

        # TODO: Validate and parse 'initial_position' and 
        # 'final_position'. These positions must be coordinates
        # on a 1920x1080 scene.
        arg = MoviepyArgument(initial_position, final_position, TFunctionSetPosition.linear, RateFunction.linear)

        return MoviepyWith.apply(video, with_position = arg)
    
    def apply_over_video(cls, video: Union[Clip, str], background_video: Union[Clip, str], initial_position: Union[Coordinate, Position], final_position: Union[Coordinate, Position]):
        arg = MoviepyArgument(initial_position, final_position, TFunctionSetPosition.linear, RateFunction.linear)

        return MoviepyWith.apply_over_video(video, background_video, with_position = arg)


from TTS.api import TTS


def narrate(text, output_filename):
    """
    @deprecated

    TODO: Remove this file and method if useless. Please, read below to check.
    This method should be removed and also the file as it is only one specific
    model in TTS narration library. It is not a different system. So please,
    remove it if it won't be used.
    """
    # TODO: Delete tortoise lib?
    # TODO: Delete en/multi-datase/tortoise-v2 model
    tts = TTS("tts_models/es/multi-dataset/tortoise-v2")
    # Check code here: https://docs.coqui.ai/en/latest/models/tortoise.html
    tts.tts_to_file(text = text, language = 'es', file_path = output_filename)

    #reference_clips = [utils.audio.load_audio(p, 22050) for p in clips_paths]
    
    #pcm_audio = tts.tts(text)
    #pcm_audio = tts.tts_with_preset("your text here", voice_samples=reference_clips, preset='fast')
    
    #from tortoise.utils.audio import load_audio, load_voice


from yta_multimedia.video.edition.modifications.video_modification import VideoModification
from yta_multimedia.video.edition.video_editor import VideoEditor
from moviepy.Clip import Clip


class ColorTemperatureVideoModification(VideoModification):
    factor: int = None
    
    def __init__(self, start_time: float, end_time: float, layer: int, factor: int = 45):
        super().__init__(start_time, end_time, layer)
        
        # TODO: Validate that 'factor' is actually between
        # the limits
        self.factor = factor

    def _modificate(self, video: Clip) -> Clip:
        return VideoEditor(video).change_color_temperature(self.factor)


from yta_multimedia.video.edition.modifications.video_modification import VideoModification
from yta_general_utils.programming.parameter_validator import PythonValidator
from moviepy.Clip import Clip


class VideoModifications:
    """
    Class to encapsulate different VideoModifications and
    to apply them in the order they must be applied.
    """
    _modifications: list[VideoModification] = None

    @property
    def modifications(self):
        """
        Return the list of 'modifications' sorted by
        layer and 'start_time' in ascending order.

        This means that the lowest layers are returned
        first, and for each 'layer' they are sorted by
        'start_time' in ascending order.
        """
        return sorted(self._modifications, key = lambda modification: (modification.layer, modification.start_time))

    # TODO: What about the type of modification? Maybe
    # I want to apply effects before greenscreens if
    # they are in the same layer, so I have to specify
    # one strategy when this happens

    def __init__(self, modifications: list[VideoModification]):
        if not PythonValidator.is_list_of(modifications, VideoModification):
            raise Exception('Not all the elements in the "modifications" parameter provided are VideoModification instances.')

        self._modifications = modifications

    def apply(self, video: Clip) -> Clip:
        for modification in self.modifications:
            previous_duration = video.duration
            
            video = modification.apply(video)

            if previous_duration != video.duration:
                # TODO: update the next modifications to
                # fit the new durations
                pass

        return video


from yta_multimedia.video.edition.resize import resize_video
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.file.checker import FileValidator

import cv2


def get_video_size(video_filename: str):
        """
        Return a tuple containing the provided
        'video_filename' size (w, h) if the parameter
        is a filename of a valid video file.
        """
        # TODO: Refactor this method to allow any kind
        # of video parameter and I detect dynamically
        # if str to use this below, or clip to use moviepy
        if not PythonValidator.is_string(video_filename):
            raise Exception('The provided "video_filename" parameter is not a valid string.')
        
        if not FileValidator.file_is_video_file(video_filename):
            raise Exception('The provided "video_filename" is not a valid video file name.')

        v = cv2.VideoCapture(video_filename)

        return int(v.get(cv2.CAP_PROP_FRAME_WIDTH)), int(v.get(cv2.CAP_PROP_FRAME_HEIGHT))

__all__ = [
    resize_video,
    get_video_size
]




from yta_general_utils.image.parser import ImageParser
from yta_general_utils.programming.parameter_validator import NumberValidator
from PIL import Image

import numpy as np
import colorsys


class ImageEditor:
    """
    Class to simplify and encapsulate all the functionality
    related to image edition.
    """
    
    @staticmethod
    def modify_color_temperature(image: any, factor: int = 0):
        # TODO: Apply 'image' typing and add to return
        return change_image_color_temperature(image, factor)

    @staticmethod
    def modify_color_hue(image: any, factor: int = 0):
        # TODO: Apply 'image' typing and add to return
        return change_image_color_hue(image, factor)



COLOR_TEMPERATURE_CHANGE_LOWER_LIMIT = -50
COLOR_TEMPERATURE_CHANGE_UPPER_LIMIT = 50

def change_image_color_temperature(image: any, factor: int = 0) -> np.ndarray:
    """
    Change the 'image' color temperature by the
    provided 'factor', that must be a value between
    [-50, 50].

    The color change consist of updating the red and
    blue values, where red is calid and blue is cold.
    Increasing the temperature means increasing the
    red color, and decreasing it, decreasing the blue
    color.
    """
    if not NumberValidator.is_number_between(factor, COLOR_TEMPERATURE_CHANGE_LOWER_LIMIT, COLOR_TEMPERATURE_CHANGE_UPPER_LIMIT):
        raise Exception(f'The "factor" parameter provided is not a number between [{COLOR_TEMPERATURE_CHANGE_LOWER_LIMIT}, {COLOR_TEMPERATURE_CHANGE_UPPER_LIMIT}].')
    
    # The '.copy()' makes it writeable
    image = ImageParser.to_numpy(image).copy()

    if factor == 0:
        return image

    # We want the factor being actually a value between 0.50 and 1.50,
    # but multiplying by 1.5 is equal to divide by 0.75 so I need to
    # manually do this calculation to apply the formula correctly
    if factor < 0:
        factor = 1 - (0.25 - normalize(factor, COLOR_TEMPERATURE_CHANGE_LOWER_LIMIT, 0, 0, 0.25))
    elif factor > 0:
        factor = 1 + normalize(factor, 0, COLOR_TEMPERATURE_CHANGE_UPPER_LIMIT, 0, 0.5)
    
    r, b = image[:, :, 0], image[:, :, 2]
    
    # Min and max values are 0 and 255
    r = np.clip(r * factor, 0, 255)
    b = np.clip(b / factor, 0, 255)
    
    # Reconstruimos la imagen con los canales modificados
    image[:, :, 0] = r
    image[:, :, 2] = b

    return image


COLOR_HUE_CHANGE_LOWER_LIMIT = -50
COLOR_HUE_CHANGE_UPPER_LIMIT = 50
# These below are 2 functions to convert
# TODO: Please, move these functions to a method
# maybe in 'yta_general_utils' or in 'yta_multimedia'
rgb_to_hsv = np.vectorize(colorsys.rgb_to_hsv)
hsv_to_rgb = np.vectorize(colorsys.hsv_to_rgb)

def change_image_color_hue(image: any, factor: int = 0) -> np.ndarray:
    """
    Change the 'image' color hue by the provided
    'factor', that must be a value between [-50, 50].

    Colorize PIL image `original` with the given
    `factor` (hue within 0-360); returns another PIL image.
    """
    if not NumberValidator.is_number_between(factor, COLOR_HUE_CHANGE_LOWER_LIMIT, COLOR_HUE_CHANGE_UPPER_LIMIT):
        raise Exception(f'The "factor" parameter provided is not a number between [{COLOR_HUE_CHANGE_LOWER_LIMIT}, {COLOR_HUE_CHANGE_UPPER_LIMIT}].')
    
    # The '.copy()' makes it writeable
    image = ImageParser.to_numpy(image).copy()
    
    factor = normalize(factor, COLOR_HUE_CHANGE_LOWER_LIMIT, COLOR_HUE_CHANGE_UPPER_LIMIT, 0, 360)
    
    # TODO: This code is not working well
    # TODO: This method is very very slow
    #arr = np.array(np.asarray(img).astype('float'))
    #r, g, b, a = np.rollaxis(image, axis = -1)
    print(image.size) # size is 6220800
    r, g, b = np.rollaxis(image, axis = -1)
    #r, g, b = np.moveaxis(image, -1, 0)
    h, s, v = rgb_to_hsv(r, g, b)
    h = factor / 360.0
    r, g, b = hsv_to_rgb(h, s, v)
    #arr = np.dstack((r, g, b, a))
    arr = np.dstack((r, g, b)).astype(np.uint8)
    print(arr.size) # size is 220800
    print(arr)

    # TODO: I don't like this line below
    return arr
    return Image.fromarray(arr.astype('uint8'), 'RGBA')


# TODO: Move this method to 'yta_general_utils' library
# and check how other normalize methods are working,
# maybe we can use this general method
def normalize(number: float, input_lower_limit: float, input_upper_limit: float, output_lower_limit: float = 0.0, output_upper_limit: float = 1.0):
    """
    Normalize the 'number' value to be between 'output_lower_limit'
    and 'output_upper_limit', according to the input provided, that
    is between the 'input_lower_limit' and 'input_upper_limit' 
    values.
    """
    if not NumberValidator.is_number(number) or not NumberValidator.is_number(input_lower_limit) or not NumberValidator.is_number(input_upper_limit) or not NumberValidator.is_number(output_lower_limit) or not NumberValidator.is_number(output_upper_limit):
        raise Exception('All the parameters must be numbers.')

    if not NumberValidator.is_number_between(number, input_lower_limit, input_upper_limit):
        raise Exception('The "number" parameter provided is not a number between the input limits provided.')
    
    if input_upper_limit <= input_lower_limit or output_upper_limit <= output_lower_limit:
        raise Exception('The upper limit must be greater than the lower limit.')
    
    return (number - input_lower_limit) / (input_upper_limit - input_lower_limit) * (output_upper_limit - output_lower_limit) + output_lower_limit


from yta_multimedia.video.generation.manim.classes.base_manim_animation import BaseManimAnimation
from yta_multimedia.video.generation.manim.classes.loading_bar.parameter_classes.loading_bar_image import LoadingBarImage
from yta_multimedia.video.generation.manim.classes.loading_bar.mobjects.loading_bar_mobject import LoadingBarMobject
from yta_multimedia.video.generation.manim.utils.dimensions import ManimDimensions
from yta_multimedia.video.generation.manim.enums import MANIM_RENDERER
from yta_multimedia.video.generation.manim.classes.base_manim_animation_wrapper import BaseManimAnimationWrapper
from yta_general_utils.programming.parameter_validator import NumberValidator
from typing import Union
from manim import *


class LoadingBarWithImagesManimAnimationWrapper(BaseManimAnimationWrapper):
    """
    Loading bar with images. It will go from 'start_percentage'
    to 'end_percentage' in the provided 'duration' time. It will
    show the bar progressing and also the 'images' if provided.

    Images will be displayed and will be static or in movement.
    If image 'start_percentage' and 'end_percentage' is the same,
    it will be static at that percentage position. If they are
    different, it will move from 'start_percentage' to
    'end_percentage' in the animation 'duration' time.

    You can put the same 'start_percentage' and 'end_percentage'
    in one image than in the loading bar to make it advance
    at the same time the progress bar does.

    Here is an example of 'images' object you can pass:
    loading_bar_images = [
        LoadingBarImage('minecraft_sword_128x128.png', 60, 0, 100),
        LoadingBarImage('arrow_up_128x128.png', -60, 0, 100),
        LoadingBarImage('minecraft_sword_128x128.png', 60, 40, 40),
        LoadingBarImage('minecraft_sword_128x128.png', 60, 78, 78)
    ]
    """
    start_percentage: float = None
    end_percentage: float = None
    duration: float = None
    images: list[LoadingBarImage] = None

    def __init__(self, start_percentage: int = 0, end_percentage: int = 100, duration: float = 5, images: list[LoadingBarImage] = None):
        # TODO: This parameter validation could be done
        # by our specific parameter validator (under
        # construction yet)
        exception_messages = []

        if not NumberValidator.is_number_between(start_percentage, 0, 100):
            exception_messages.append('The "start_percentage" parameter must be a valid int between [0, 100]')
        
        if not NumberValidator.is_number_between(end_percentage, 0, 100):
            raise Exception('The "end_percentage" parameter must be a valid int between [0, 100]')
        
        if not NumberValidator.is_number_between(duration, 0, 120, do_include_lower_limit = False):
            raise Exception('"duration" parameter must be a valid float number between (0, 120]')
        
        if len(exception_messages) > 0:
            raise Exception('\n'.join([exception_message for exception_message in exception_messages]))
        
        images = [image.toJSON() for image in images] if images and len(images) > 0 else []
        
        self.start_percentage = start_percentage
        self.end_percentage = end_percentage
        self.duration = duration
        self.images = images
        super().__init__(LoadingBarWithImagesManimAnimationGenerator)

class LoadingBarWithImagesManimAnimationGenerator(BaseManimAnimation):
    """
    Loading bar with images. It will go from 'start_percentage'
    to 'end_percentage' in the provided 'duration' time. It will
    show the bar progressing and also the 'images' if provided.

    Images will be displayed and will be static or in movement.
    If image 'start_percentage' and 'end_percentage' is the same,
    it will be static at that percentage position. If they are
    different, it will move from 'start_percentage' to
    'end_percentage' in the animation 'duration' time.

    You can put the same 'start_percentage' and 'end_percentage'
    in one image than in the loading bar to make it advance
    at the same time the progress bar does.

    Here is an example of 'images' object you can pass:
    loading_bar_images = [
        LoadingBarImage('minecraft_sword_128x128.png', 60, 0, 100),
        LoadingBarImage('arrow_up_128x128.png', -60, 0, 100),
        LoadingBarImage('minecraft_sword_128x128.png', 60, 40, 40),
        LoadingBarImage('minecraft_sword_128x128.png', 60, 78, 78)
    ]
    """
    def construct(self):
        """
        This method is called by manim when executed by shell and
        will call the scene animation render method to be processed
        and generated.
        """
        self.animate()

    def generate(self, parameters: dict, output_filename: Union[str, None] = None):
        """
        Build the manim animation video and stores it
        locally as the provided 'output_filename', 
        returning the final output filename.

        This method will call the '.animate()' method
        to build the scene with the instructions that
        are written in that method.
        """
        return super().generate(
            parameters,
            renderer = MANIM_RENDERER.CAIRO,
            output_filename = output_filename
        )
    
    def animate(self):
        loading_bar = LoadingBarMobject()
        
        # We make LoadingBarImages objects again
        images = []
        for image in self.parameters['images']:
            images.append(LoadingBarImage(image['image_filename'], ManimDimensions.manim_height_to_height(image['y']), image['start_percentage'], image['end_percentage']))

        loading_bar_animation = loading_bar.get_animation(images, self.parameters['start_percentage'], self.parameters['end_percentage'], self.parameters['duration'])

        # We add the mobjects to forze them to appear
        self.add(*loading_bar.get_mobjects())

        self.play(*loading_bar_animation, rate_func = linear, run_time = self.parameters['duration'])


from yta_multimedia.video import MPVideo


class VideoFilter:
    # TODO: Move these methods to a more specific file
    # or class
    
    @staticmethod
    def apply_t_effect(frame, t: float, fps: int, frames_indexes: list[int], effect: callable, *args):
        if MPVideo.frame_time_to_frame_index(t, fps) in frames_indexes:
            return effect(frame, *args)

        return frame


def test():
    # TODO: This below has been used in 'youtube-stuff' to
    # test a new way of handling frame effects
    # TODO: 'video' is just a video
    from yta_multimedia.image.edition import ImageFilter
    from yta_multimedia.image.edition import MotionBlurDirection
    from random import choice

    def image_filter_arguments(index: int):
        # TODO: Of course, index must be ok
        # TODO: This can be whatever I want, maybe a class
        # in which I push my arrays in execution time and
        # have the logic in code

        # We have the effect values ordered in the same 
        # order as per each frame, so we return them to 
        # be applied like that
        effect_values = [10, 20, 30, 40, 30, 20, 10, 20, 10, 80]

        return effect_values[index], choice(MotionBlurDirection.get_all())

    frames_indexes = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    from moviepy.Clip import Clip
    from typing import Union

    class FramesEffect:
        frames_indexes: list[int] = None
        method: callable = None
        args_generator_function: callable = None

        def __init__(self, frames_indexes: list[int], method: callable, args_generator_function: callable):
            self.frames_indexes = frames_indexes
            self.method = method
            self.args_generator_function = args_generator_function

        def validate(self):
            # TODO: The length of all args returned by
            # 'args_generator_function' must be equal to
            # 'frames_indexes' length

            # TODO: Raise Exception if invalid
            return True
        
        def apply(self, video: Clip, output_filename: Union[str, None] = None):
            def apply_effect_on_frame(frame, t: float):
                try:
                    index = self.frames_indexes.index(MPVideo.frame_time_to_frame_index(t, video.fps))
                    print(f'Hi, I am in {index}')

                    return self.method(frame, *self.args_generator_function(index))
                except:
                    pass

                return frame
                                        
            video = video.transform(lambda get_frame, t: apply_effect_on_frame(get_frame(t), t))

            if output_filename is not None:
                video.write_videofile(output_filename)

            return video
        

    # TODO: Think about how to make the method that
    # returns the arguments be easier to create and
    # apply, because it is not easy to understand...
    FramesEffect(frames_indexes, ImageFilter.motion_blur, image_filter_arguments).apply(FramesEffect(frames_indexes, ImageFilter.motion_blur, image_filter_arguments).apply(video)).write_videofile('a_a_borrame.mp4')


"""
# Example of a function using those frame numbers
# or indexes in a condition to apply or not an
# effect to a frame
apply_motion_blur_at_begining(frame, t, motion_blur_kernels, fps):
    frame_number = MPVideo.frame_time_to_frame_index(t, fps)
    
    if frame_number < len(motion_blur_kernels):
        return ImageFilter.motion_blur(frame, motion_blur_kernels[frame_number])
    
    return frame
"""


from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.subclip import subclip_video
from yta_general_utils.programming.parameter_validator import NumberValidator, PythonValidator
from moviepy.Effect import Effect
from moviepy.Clip import Clip
from moviepy import concatenate_videoclips
from abc import abstractmethod


class MEffect(Effect):
    """
    Effect to be applied on a single video without any
    other video dependence.

    This effect cannot be applied as moviepy effects as
    the structure is different.

    TODO: If I want to make my own moviepy effects that
    are able to be applied as moviepy does, I need to
    respect the structure and just manipulate the frames.
    So, those new custom moviepy effects sohuld be placed
    in other folder as this MEffect is my own effect that
    works in a different way.
    """
    # TODO: Is this working (?)
    result_must_replace: bool = False
    """
    This parameter indicates if this effect, when applied,
    must replace the original clip part or, if False, must
    be concatenated.
    """
    @abstractmethod
    def apply_over_video(self, video: Clip, background_video: Clip):
        # TODO: Should I (?)
        pass

    @staticmethod
    def apply_effect_partially_on_video(video: Clip, effect: 'MEffect', start_time: float = 0, end_time: float = None):
        """
        Apply the provided 'effect' in the also provided 'video'
        in the part between the given 'start_time' and 'end_time'.
        """
        video = VideoParser.to_moviepy(video)

        # TODO: It will be a subclass of it, does it work (?)
        if not PythonValidator.is_instance(effect, MEffect):
            raise Exception('The provided "effect" parameter is not a valid effect. It must be an instance of MEffect class.')

        if not NumberValidator.is_number_between(start_time, 0, video.duration) or not NumberValidator.is_number_between(end_time, 0, video.duration) or start_time >= end_time:
            raise Exception('The provided "start_time" and "end_time" parameters ("{str(start_time)}" and "{str(end_time)}") must be values between 0 and the provided clip duration ("{str(video.duration)}"), and "end_time" must be greater than "start_time".')
        
        left, center, right = subclip_video(video, start_time, end_time)
        # TODO: What if only '.apply()' instead of '.apply_over_video()'
        # is needed (?)
        center = effect.apply_over_video(center.with_subclip(start_time, end_time))

        clips = [clip for clip in [left, center, right] if clip is not None]

        return concatenate_videoclips(clips)#.with_start(start)


from yta_general_utils.programming.parameter_validator import NumberValidator
from yta_multimedia.video.edition.subclip import subclip_video
from moviepy.Clip import Clip
from moviepy import concatenate_videoclips
from abc import ABC, abstractmethod


class VideoModification(ABC):
    """
    A modification that must be applied in a video. This
    class must be implemented by the specific video
    modifications.
    """
    start_time: float = None
    end_time: float = None
    layer: int = None
    
    def __init__(self, start_time: float, end_time: float, layer: int):
        if not NumberValidator.is_positive_number(start_time, do_include_zero = True):
            raise Exception('The provided "start_time" parameter is not a positive number.')
        
        if not NumberValidator.is_positive_number(end_time, do_include_zero = True):
            raise Exception('The provided "end_time" parameter is not a positive number.')
        
        if end_time <= start_time:
            raise Exception('The provided "end_time" is before the also provided "start_time".')
        
        if not NumberValidator.is_positive_number(layer, do_include_zero = True):
            raise Exception('The provided "layer" parameter is not a positive number.')
        
        self.start_time = start_time
        self.end_time = end_time
        self.layer = layer

    def apply(self, video: Clip):
        """
        Apply the modification making the necessary subclips.
        This method will apply the modification that must be
        set in the 'apply' method in the subclass subclipping
        the necessary according to the 'start_time' and
        'end_time'.
        """
        if self.start_time > video.duration or self.end_time > video.duration:
            raise Exception('This VideoModification cannot be applied in the provided "video".')

        left_clip, center_clip, right_clip = subclip_video(video, self.start_time, self.end_time)
        center_clip = self._modificate(center_clip)

        clips = [clip for clip in [left_clip, center_clip, right_clip] if clip is not None]

        return concatenate_videoclips(clips)

    @abstractmethod
    def _modificate(self, video: Clip):
        """
        Method to be specifically set in each modification
        with the custom video modification code. This method
        will generate a video that will replace the original
        'video' provided.

        Do not call this method, call the '.apply()' method
        to actually apply it.
        """
        pass


from yta_multimedia.audio.converter import AudioConverter
from moviepy import AudioClip
from pydub import AudioSegment
from typing import Union

import numpy as np


class AudioParser:
    """
    Class to simplify the way we parse audios.
    """
    @staticmethod
    def as_audioclip(audio: Union[str, np.ndarray, AudioSegment, AudioClip]):
        audio, _ = AudioConverter.to_audioclip(audio)

        return audio
    
    @staticmethod
    def as_audiosegment(audio: Union[str, np.ndarray, AudioSegment, AudioClip]):
        audio, _ = AudioConverter.to_audiosegment(audio)

        return audio
    
    # TODO: '.as_numpy()' ? It is difficult due to rate
    # or strange mapping... (?)


from yta_multimedia.video.consts import MOVIEPY_SCENE_DEFAULT_SIZE, MANIM_SCENE_DEFAULT_SIZE, DEFAULT_SCENE_SIZE
from yta_general_utils.coordinate.coordinate import Coordinate as BaseCoordinate
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.parameter_validator import PythonValidator
from typing import Union
from copy import copy


class VideoEngine(Enum):
    """
    Enum class to represent the different video engine
    systems we are using in this app.
    """
    MOVIEPY = 'moviepy'
    """
    The moviepy video engine which uses the upper left
    corner as the origin of the movement and has a scene
    which origin (0, 0) is the upper left corner, being
    positive values bottom and right directions and 
    negative top and left.
    """
    MANIM = 'manim'
    """
    The manim video engine which uses the center of the
    video as the origin of the movement and has a scene
    which origin (0, 0) is in the middle being positive
    values top and right directions and negative bottom
    and left.
    """

class Coordinate(BaseCoordinate):
    """
    Class that represents a position in a scene with
    of specific dimensions. This class will always
    refer to the center of the element positioned in
    that position.

    This class contains method to calculate the corner
    of the element to be able to move moviepy videos,
    but it is always pointing to the center of the 
    element.
    """
    engine: VideoEngine = None
    """
    The video engine for which this coordinate has
    been created.
    """
    scene_size: tuple = None
    """
    The scene dimensions (for moviepy engine) the
    coordinate has been built for.
    """

    def __init__(self, x: float, y: float, is_normalized: bool = False, engine: Union[VideoEngine, str] = VideoEngine.MOVIEPY, scene_size: tuple = MOVIEPY_SCENE_DEFAULT_SIZE):
        super().__init__(x, y, is_normalized)
        engine = VideoEngine.to_enum(engine)

        if scene_size is None:
            scene_size = MOVIEPY_SCENE_DEFAULT_SIZE
        else:
            super().validate(scene_size, 'scene_size')
        
        self.type = type
        self.engine = engine
        self.scene_size = scene_size

    def to_moviepy(self):
        """
        This method updates the instance 'x', 'y' and 'engine'
        values if the current engine is Manim. If this happens,
        the scene will be set as (1920, 1080), as this is the
        default scene we consider for manim.
        """
        if self.engine == VideoEngine.MANIM:
            new_coordinate = self.as_moviepy_tuple()

            self.x = new_coordinate[0]
            self.y = new_coordinate[1]
            # As we consider our manim scene similar to one
            # moviepy 1920x1080 scene, we force it
            self.scene_size = MOVIEPY_SCENE_DEFAULT_SIZE
            self.engine = VideoEngine.MOVIEPY

        return self
    
    def to_manim(self):
        """
        This method updates the instance 'x', 'y' and 'engine'
        values if the current engine is Moviepy. If this happens,
        the scene will be set as (1920, 1080), as this is the
        default scene we consider for moviepy.
        """
        if self.engine == VideoEngine.MOVIEPY:
            new_coordinate = self.as_manim_tuple()

            self.x = new_coordinate[0]
            self.y = new_coordinate[1]
            # As we consider our manim scene similar to one
            # moviepy 1920x1080 scene, we force it
            self.scene_size = MOVIEPY_SCENE_DEFAULT_SIZE
            self.engine = VideoEngine.MANIM

        return self

    def as_moviepy_tuple(self):
        """
        Return this Coordinate instance as a moviepy coordinate
        tuple. This method will return a value but not update
        any instance attribute.

        This method forces the scene to be (1920, 1080) that
        is the default scene size (in normal size terms).
        """
        x, y = self.x, self.y
        if self.engine == VideoEngine.MANIM:
            # As we consider our manim scene similar to one
            # moviepy 1920x1080 scene, we need to make sure
            # it is refering that screen size.
            # TODO: I need to refactor this. When 'x' is 0
            # in moviey in a 1920x1080 scene, the manim value
            # would be -
            x, y = self.on_scene_size(MOVIEPY_SCENE_DEFAULT_SIZE)
            x, y = Coordinate.manim_coordinate_to_moviepy_coordinate((x, y))

        return x, y
    
    def as_manim_tuple(self):
        """
        Return this Coordinate instance as a manim coordinate
        tuple. This method will return a value but not update
        any instance attribute.

        This method forces the scene to be (1920, 1080) that
        is the default scene size (in normal size terms).
        """
        x, y = self.x, self.y
        if self.engine == VideoEngine.MOVIEPY:
            # As we consider our manim scene similar to one
            # moviepy 1920x1080 scene, we need to make sure
            # it is refering that screen size.
            x, y = self.on_scene_size(MOVIEPY_SCENE_DEFAULT_SIZE)
            x, y = Coordinate.moviepy_coordinate_to_manim_coordinate((x, y))

        return x, y

    def update_scene_size(self, scene_size: tuple = MOVIEPY_SCENE_DEFAULT_SIZE):
        """
        Update the current scene size and also the coordinate
        values to fit the new scene size. This method updates
        this Coordiante instance attributes.
        """
        x, y = self.on_scene_size(scene_size)

        # TODO: What if Manim (?)
        self.x = x
        self.y = y
        self.scene_size = scene_size

        return self
    
    def on_scene_size(self, scene_size: tuple = MOVIEPY_SCENE_DEFAULT_SIZE):
        """
        Obtain the 'x' and 'y' values that fit the provided 
        'scene_size' by making the required calculations.

        This method do not modify the instance.
        """
        super().validate(scene_size, 'scene_size')

        x, y = self.x, self.y
        if self.scene_size != scene_size:
            tmp: Coordinate = copy(self)
            tmp = tmp.update_scene_size(scene_size)
            x, y = tmp.x, tmp.y
        
        return x, y

    def get_moviepy_upper_left_corner_tuple(self, video_size: tuple):
        """
        Calculate the upper left corner tuple of this Coordinate
        instance based on its screen_size and the provided
        'video_size'.
        """
        super().validate(video_size, 'video_size')

        # TODO: What if Manim (?)
        if self.engine == VideoEngine.MANIM:
            raise Exception('Please, write the code for this.')

        return self.x - video_size[0] / 2, self.y - video_size[1] / 2
    
    @staticmethod
    def to_moviepy_upper_left_corner_tuple(coordinate: Union[tuple, 'Coordinate'], video_size: Union[tuple, 'Coordinate'], scene_size: Union[tuple, 'Coordinate'] = MOVIEPY_SCENE_DEFAULT_SIZE):
        """
        Calculate the upper left corner using the provided
        'coordinate', 'video_size' and making the calculations
        based on a scene of the provided 'scene_size'.
        """
        super().validate(coordinate, 'coordinate')
        super().validate(video_size, 'video_size')
        super().validate(scene_size, 'scene_size')

        if PythonValidator.is_instance(coordinate, Coordinate):
            # TODO: What about manim (?)
            # TODO: Is this updating the reference?, If yes,
            # that is an unexpected behaviour
            #coordinate = coordinate.update_scene_size(scene_size)
            tmp: Coordinate = copy(coordinate)
            tmp = tmp.update_scene_size(scene_size)
        else:
            coordinate = Coordinate(coordinate[0], coordinate[1], scene_size = scene_size)

        return coordinate.get_moviepy_upper_left_corner_tuple(video_size)
    
    @staticmethod
    def moviepy_coordinate_to_manim_coordinate(coordinate: tuple):
        """
        Transform the provided moviepy 'coordinate' to
        a manim coordinate. Each of those coordinates 
        must be considered as built in a 1920x1080 scene.

        Please, ensure the provided 'coordinate' is 
        representing a point within a scene of 1920x1080.
        """
        super().validate(coordinate)

        # Scale for moviepy -> manim conversion
        x_scale = MANIM_SCENE_DEFAULT_SIZE[0] / DEFAULT_SCENE_SIZE[0]
        y_scale = MANIM_SCENE_DEFAULT_SIZE[1] / DEFAULT_SCENE_SIZE[1]
        
        x = (coordinate[0] - DEFAULT_SCENE_SIZE[0] / 2) * x_scale
        y = (DEFAULT_SCENE_SIZE[1] / 2 - coordinate[1]) * y_scale

        return x, y
    
    @staticmethod
    def manim_coordinate_to_moviepy_coordinate(coordinate: tuple):
        """
        Transform the provided manim 'coordinate' to
        a moviepy coordinate. Each of those coordinates 
        must be considered as built in a 1920x1080 scene.

        Please, ensure the provided 'coordinate' is 
        representing a point within a scene of 1920x1080.
        """
        super().validate(coordinate)

        # Scale for manim -> moviepy conversion
        x_scale = DEFAULT_SCENE_SIZE[0] / MANIM_SCENE_DEFAULT_SIZE[0]
        y_scale = DEFAULT_SCENE_SIZE[1] / MANIM_SCENE_DEFAULT_SIZE[1]
        
        x = (coordinate[0] * x_scale) + MANIM_SCENE_DEFAULT_SIZE[0] / 2
        y = MANIM_SCENE_DEFAULT_SIZE[1] / 2 - (coordinate[1] * y_scale)

        return x, y

    @staticmethod
    def to_coordinate(coordinate: tuple):
        """
        Turn the provided tuple into a Coordinate instance.
        """
        super().validate(coordinate)

        return Coordinate(coordinate[0], coordinate[1])


"""
When we need to use videos generated with manim
we have many different types of videos, and we
need to ensure that the provided wrapper class
is one of the types the method we are using is
expecting.

If we are trying to overlaying a text which is
generated with a text manim wrapper class, we
need to raise an exception if the provided class
is not a text manim wrapper class, because the
process will fail as the video generated will be
different as the expected.

All the classes we have that belong to manim video
creation have the same structure, having a wrapper
class that internally uses a generator class to
actually build the video animation, so we need
those wrapper class names. But also, the wrapper
class name is the same as the file name but in
camel case and ending in 'Wrapper'.
"""
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.generation.manim.classes.base_manim_animation_wrapper import BaseManimAnimationWrapper, ManimAnimationType
from yta_multimedia.video.edition.resize import resize_video
from yta_multimedia.video.edition.duration import set_video_duration, ExtendVideoMode
from yta_multimedia.video.combine import VideoCombinatorAudioMode, VideoAudioCombinator
from yta_multimedia.video.edition.subclip import subclip_video
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_general_utils.programming.parameter_validator import PythonValidator, NumberValidator
from yta_general_utils.file.handler import FileSearchOption, FileHandler
from yta_general_utils.programming.var import snake_case_to_upper_camel_case
from yta_general_utils.programming.path import get_project_abspath
from moviepy.Clip import Clip
from moviepy import CompositeVideoClip, concatenate_videoclips
from abc import ABC, abstractmethod


# TODO: Check yta_multimedia\video\combine.py because
# its functionality could end being part of this 
# VideoEditor


# TODO: Please, rename this class as this name is
# not a proper name
class VideoClassifier:
    # TODO: This method is no longer being used I think
    @staticmethod
    def get_manim_wrapper_class_names_from_files(abspath: str, files_to_ignore: list[str] = []):
        """
        Obtain a list with the manim wrapper class names of
        all the available files that are in the provided
        'abspath', excluding the ones in the also given
        'files_to_ignore'. The file name is turned into the
        wrapper class name and returned.
        """
        files_to_ignore = [files_to_ignore] if PythonValidator.is_string(files_to_ignore) else files_to_ignore

        if not PythonValidator.is_list_of_string(files_to_ignore):
            raise Exception('The "files_to_ignore" parameter provided is not a valid list of strings.')

        # Transform the file name in the wrapper class that is inside
        transform_function = lambda file: snake_case_to_upper_camel_case(file.split("/")[-1].replace(".py", ""))

        return [
            f'{transform_function(file)}Wrapper'
            for file in FileHandler.get_list(abspath, FileSearchOption.FILES_ONLY, '*.py')
            if not any(file.endswith(file_to_ignore) for file_to_ignore in files_to_ignore)
        ]

        # TODO: Maybe try another way of getting all the classes
        # within a module, not a file, and identify like I tried
        # with this 'get_manim_wrapper_class_names_from_files'
        # method that is not working because files change when
        # imported as library
        return VideoClassifier.get_manim_wrapper_class_names_from_files(
            f'{get_project_abspath()}/video/generation/manim/classes/text/',
            ['__init__.py']
        )

SIZE_FACTOR = 4

class VideoEditor:
    _video: Clip = None

    @property
    def video(self):
        return self._video

    def __init__(self, video: Clip):
        self._video = VideoParser.to_moviepy(video, do_include_mask = True, do_calculate_real_duration = True)

    def overlay_text(self, text_generator_wrapping_instance: BaseManimAnimationWrapper):
        self._validate_is_manim_wrapper_instance_of_type(text_generator_wrapping_instance, ManimAnimationType.TEXT_ALPHA)
        
        video = VideoParser.to_moviepy(text_generator_wrapping_instance.generate(), do_include_mask = True)
        video = _prepare_video(self.video, video, 1)
        video = _overlay_video(self.video, video, position = ('center', 'center'), audio_mode = VideoCombinatorAudioMode.ONLY_MAIN_CLIP_AUDIO)

        return video
    
    def overlay_video_without_alpha_fullscreen(self, video: Clip, audio_mode: VideoCombinatorAudioMode = VideoCombinatorAudioMode.BOTH_CLIPS_AUDIO):
        """
        Useful to show a stock video while the main clip is
        still speaking, or to focus on the stock video.
        """
        video = VideoParser.to_moviepy(video)
        audio_mode = VideoCombinatorAudioMode.to_enum(audio_mode)

        video = _prepare_video(self.video, video, 1)
        video = _overlay_video(self.video, video, position = ('center', 'center'), audio_mode = audio_mode)

        return video

    def overlay_video_without_alpha_non_fullscreen(self, video: Clip, audio_mode: VideoCombinatorAudioMode = VideoCombinatorAudioMode.BOTH_CLIPS_AUDIO):
        """
        Useful to add a video like a reel or stock while the
        main clip is still visible.
        """
        video = VideoParser.to_moviepy(video)
        audio_mode = VideoCombinatorAudioMode.to_enum(audio_mode)

        video = _prepare_video(self.video, video, SIZE_FACTOR)
        video = _overlay_video(self.video, video, position = ('center', 'center'), audio_mode = audio_mode)

        return video
    
    def overlay_video_with_alpha_fullscreen(self, video: Clip, audio_mode: VideoCombinatorAudioMode = VideoCombinatorAudioMode.BOTH_CLIPS_AUDIO):
        """
        Useful to add an alphascreen, a transition or
        another kind of videos.
        """
        video = VideoParser.to_moviepy(video, do_include_mask = True)
        audio_mode = VideoCombinatorAudioMode.to_enum(audio_mode)

        video = _prepare_video(self.video, video, 1)
        video = _overlay_video(self.video, video, position = ('center', 'center'), audio_mode = audio_mode)

        return video
    
    def overlay_video_with_alpha_non_fullscreen(self, video: Clip, audio_mode: VideoCombinatorAudioMode = VideoCombinatorAudioMode.BOTH_CLIPS_AUDIO):
        """
        Useful for something that I don't know right now.

        TODO: Please, improve this doc... omg
        """
        video = VideoParser.to_moviepy(video, do_include_mask = True)
        audio_mode = VideoCombinatorAudioMode.to_enum(audio_mode)

        video = _prepare_video(self.video, video, SIZE_FACTOR)
        video = _overlay_video(self.video, video, position = ('center', 'center'), audio_mode = audio_mode)

        return video

    # Basic effects imitating capcut here below
    def zoom(self, factor: int = 100):
        """
        Apply a zoom in the video. 1% means zooming out to 1/100 of
        the video size, while 500% means zooming in to 5 times its 
        size.

        TODO: The 'zoom' method name is not very self-descriptive
        """
        # TODO: Any limit must be set in a general VideoEditor
        # settings file
        ZOOM_LIMIT = (1, 500)
        
        if not NumberValidator.is_number_between(ZOOM_LIMIT[0], ZOOM_LIMIT[1]):
            raise Exception(f'The "factor" parameter provided is not a number between [{ZOOM_LIMIT[0]}, {ZOOM_LIMIT[1]}].')
        
        # We apply a black background to ensure the video size
        # is the expected one and no problems with resizing
        black_background = ClipGenerator.get_default_background_video(duration = self.video.duration)
        new_size = (factor / 100 * self.video.size[0], factor / 100 * self.video.size[1])

        return _overlay_video(black_background, self.video.resized(new_size))
    
    def move(self, x_variation: int = 0, y_variation: int = 0):
        """
        Apply a movement in the video, which means that it will be 
        not centered if 'x_variation' and/or 'y_variation' are 
        different from zero.

        TODO: I don't like the 'move' method name
        """
        # TODO: Any limit must be set in a general VideoEditor
        # settings file
        X_LIMIT = (-1920, 1920)
        Y_LIMIT = (-1080, 1080)

        if not NumberValidator.is_number_between(x_variation, X_LIMIT[0], X_LIMIT[1]):
            raise Exception(f'The provided "x_variation" must be between [{X_LIMIT[0]}, {X_LIMIT[1]}].')
        
        if not NumberValidator.is_number_between(y_variation, Y_LIMIT[0], Y_LIMIT[1]):
            raise Exception(f'The provided "y_variation" must be between [{Y_LIMIT[0]}, {Y_LIMIT[1]}]')
        
        # We apply a black background to ensure the video size
        # is the expected one and no problems with movement
        black_background = ClipGenerator.get_default_background_video(duration = self.video.duration)

        return _overlay_video(black_background, self.video, position = (x_variation, y_variation))
        
    def change_color_temperature(self, factor: int = 0):
        from yta_multimedia.image.edition.image_editor import COLOR_TEMPERATURE_CHANGE_LOWER_LIMIT, COLOR_TEMPERATURE_CHANGE_UPPER_LIMIT, ImageEditor

        if not NumberValidator.is_number_between(factor, COLOR_TEMPERATURE_CHANGE_LOWER_LIMIT, COLOR_TEMPERATURE_CHANGE_UPPER_LIMIT):
            raise Exception(f'The "factor" parameter provided is not a number between [{COLOR_TEMPERATURE_CHANGE_LOWER_LIMIT}, {COLOR_TEMPERATURE_CHANGE_UPPER_LIMIT}].')
        
        # TODO: Do I need to copy() (?)
        return self.video.transform(
            lambda get_frame, t:
            ImageEditor.modify_color_temperature(get_frame(t), factor)
        )
    
    def change_color_hue(self, factor: int = 0):
        from yta_multimedia.image.edition.image_editor import COLOR_HUE_CHANGE_LOWER_LIMIT, COLOR_HUE_CHANGE_UPPER_LIMIT, ImageEditor

        if not NumberValidator.is_number_between(factor, COLOR_HUE_CHANGE_LOWER_LIMIT, COLOR_HUE_CHANGE_UPPER_LIMIT):
            raise Exception(f'The "factor" parameter provided is not a number between [{COLOR_HUE_CHANGE_LOWER_LIMIT}, {COLOR_HUE_CHANGE_UPPER_LIMIT}].')
        
        # TODO: Do I need to copy() (?)
        return self.video.transform(
            lambda get_frame, t:
            ImageEditor.modify_color_hue(get_frame(t), factor)
        )
    
    # TODO: Move this method to another place maybe
    def _validate_is_manim_wrapper_instance_of_type(self, manim_wrapper_instance: BaseManimAnimationWrapper, type: ManimAnimationType):
        """
        Validate that the provided 'manim_wrapper_instance'
        is actually an instance of that class and also of
        the given 'type'.
        """
        # The instance must be an instance of the specific
        # class that identifies it as a manim animation
        # generator wrapper.
        # This class contains the parameters needed (with
        # their values actually set) and the animation
        # generator class that must be called with those
        # parameters to generate the animation video.
        if not PythonValidator.is_subclass(manim_wrapper_instance, BaseManimAnimationWrapper) or not PythonValidator.is_an_instance(manim_wrapper_instance):
            raise Exception('The "text_generator_wrapping_instance" is not a valid instance of a subclass of BaseManimAnimationWrapper class.')
        
        type = ManimAnimationType.to_enum(type)

        # Validate the wrapping instance is of the expected
        # type
        if type in manim_wrapper_instance.types:
            raise Exception('The provided "text_generator_wrapping_instance" is not an instance of a manim text generation class.')

    

def _prepare_video(main_video: Clip, video: Clip, size_factor: float = 1.0):
    """
    Resize the 'video' according to the 'main_video' dimensions
    and enshort the 'video' if larger than the 'main_video'.
    """
    # We resize the 'video' to fit expected size
    video = resize_video(video, tuple(size_element / size_factor for size_element in main_video.size))
    # We ensure the video is not larger than the main one
    video = set_video_duration(video, main_video.duration, extend_mode = ExtendVideoMode.DONT_ENLARGE)

    return video

def _overlay_video(background_video: Clip, video: Clip, position: tuple = ('center', 'center'), audio_mode: VideoCombinatorAudioMode = VideoCombinatorAudioMode.BOTH_CLIPS_AUDIO):
    """
    Overlay the provided 'video' on top of the also given
    'background_video'.
    """
    return CompositeVideoClip([
        background_video,
        video.with_position(position)
    ]).with_audio(VideoAudioCombinator(audio_mode).process_audio(background_video, video))





# TODO: Move this to a better place
    
# TODO: I'm creating this raw class to use as a valid
# and working example of what I want to have in code
# and to apply, so later I can think about the best
# structure and hierarchy to allow it



class OverlayTextVideoModification(VideoModification):
    """
    Simple class that represents a video modification that
    consist of a text that shown over a video.
    """
    text: str = None
    start_time: float = None
    end_time: float = None
    generator_class: any = None

    def __init__(self, text: str, start_time: float):
        # TODO: Add all parameters and validate them
        # TODO: This is an specific effect that will use a manim
        # wrapper class to build the text that will be overlayed
        # so I'm not sure how to handle this (inherit, accept the
        # wrapper class as parameter, etc.)
        pass

    def apply(self, video: Clip) -> Clip:
        # TODO: Generate the text and apply it
        return video

"""
So, we have an 'OverlayTextVideoModification' which is an
specific modification that consis of adding a text in overlay
mode. This is a 'VideoModification', so it will be accepted
as a VideoModification valid class to apply. We can apply it.
"""



# 1. Videos must be 60fps both of them to simplify
# 2. The main video (background_video) must be 1920x1080 always,
#    and the other ones must be 1920x1080 or smaller
# 3. Duration of the video cannot be larger than the main video

# We should add a VideoModifications matrix in which we have
# layers that indicate the moment in which the modification has
# to be applied. Layer 1 will be prior, so once all layer 1
# modifications has been completed, layer 2 are applied. This
# is how editors work and also the better way to handle 
# priority. It is not the same applying a greenscreen and then
# an effect than applying the effect first to the clip and then
# the greenscreen that wraps the whole video.


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.position import Position
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_multimedia.video.edition.effect.moviepy.position.objects.coordinate import Coordinate
from yta_multimedia.video.edition.effect.moviepy.position.move.move_linear_position_effect import MoveLinearPositionEffect
from yta_multimedia.video.edition.effect.moviepy.position.slide import get_in_and_out_positions_as_list
from moviepy.Clip import Clip
from moviepy import concatenate_videoclips
from typing import Union


class SlideInAndOutRandomlyEffect(Effect):
    """
    Slides from outside the screen to the specified position
    (which is the center by default), stays there and goes
    away through the opposite side.
    """
    def apply(self, video: Clip, position: Union[Position, Coordinate, tuple] = Position.CENTER) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlurEffect.apply, locals(), ['video'])
        background_video = ClipGenerator.get_default_background_video(duration = video.duration)

        return self.apply_over_video(video, background_video, position)
    
    def apply_over_video(self, video: Clip, background_video: Clip, position: Union[Position, Coordinate] = Position.CENTER) -> Clip:
        random_position = get_in_and_out_positions_as_list()

        # TODO: Is this ok (?)
        # video_handler = MPVideo(video)
        # background_video = video_handler.prepare_background_clip(background_video)

        movement_time = background_video.duration / 6
        stay_time = background_video.duration / 6 * 4

        effect = concatenate_videoclips([   
            MoveLinearPositionEffect().apply_over_video(
                video.with_subclip(0, movement_time),
                background_video.with_subclip(0, movement_time),
                random_position[0],
                position
            ),
            # TODO: This can be replaced by a StayAtPositionEffect
            # but the result is the same actually
            MoveLinearPositionEffect().apply_over_video(
                video.with_subclip(movement_time, movement_time + stay_time),
                background_video.with_subclip(movement_time, movement_time + stay_time),
                position,
                position
            ),
            MoveLinearPositionEffect().apply_over_video(
                video.with_subclip(movement_time + stay_time, video.duration),
                background_video.with_subclip(movement_time + stay_time, video.duration),
                position,
                random_position[1]
            )
        ])

        return effect


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames.video_frame_extractor import VideoFrameExtractor
from moviepy import Clip, ImageClip, concatenate_videoclips


class StopMotionEffect(Effect):
    """
    Creates a Stop Motion effect in the provided video
    by dropping the frames per second but maintaining
    the original frames ratio.
    """
    def apply(self, video: Clip) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlinkEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        FRAMES_TO_JUMP = 5

        clips = []
        for frame_number in range((int) (video.fps * video.duration)):
            if frame_number % FRAMES_TO_JUMP == 0:
                frame = VideoFrameExtractor.get_frame_by_number(video, frame_number)
                # TODO: What about the mask (?)
                clips.append(ImageClip(frame, duration = FRAMES_TO_JUMP / video.fps).with_fps(video.fps))

        return concatenate_videoclips(clips).with_audio(video.audio).with_fps(video.fps)


from yta_multimedia.video.frames.numpy_frame_helper import NumpyFrameHelper, FrameMaskingMethod

import numpy as np


class VideoFrame:
    """
    Class to represent a video frame, to simplify the way
    we turn it into a mask and work with it. A mask frame
    is a numpy array with single values per pixel and 
    values between 0.0 and 1.0 (where 0.0 is full 
    full transparent and 1.0 is full opaque). A normal
    frame is a numpy array with 3 values per pixels
    (representing R, G, B) and values between 0 and 255 
    (where 0 is the absence of color and 255 the full
    presence of that color).

    This class has been created to work easily with frames
    and to verify them when creating a new instance.
    """
    frame: np.ndarray = None
    """
    The frame information as a numpy array. This array can
    only contain frames in the format of not normalized
    RGB (array of 3 values from 0 to 255 per pixel) or
    normalized alpha (1 single value per pixel from 0.0 to
    1.0).
    """
    _is_mask: bool = None

    @property
    def is_mask(self):
        """
        Return True if the stored 'frame' is a mask, which means
        that it is a numpy array of single values between 0.0 and
        1.0 (representing the opacity, which 0.0 as full
        transparent and 1.0 as full opaque).
        """
        return self._is_mask
    
    @property
    def is_normal(self):
        """
        Return True if the stored 'frame' is a normal frame, which
        means that it is a numpy array of 3 values (R, G, B) between
        0 and 255 (representing the color presence, with 0 as no
        color and 255 as full color).
        """
        return self._is_mask == False

    def __init__(self, frame: np.ndarray):
        if not NumpyFrameHelper.is_rgb_not_normalized(frame) and not NumpyFrameHelper.is_rgb_normalized(frame) and not NumpyFrameHelper.is_alpha_normalized(frame) and not NumpyFrameHelper.is_alpha_not_normalized(frame):
            # TODO: Print properties to know why it is not valid
            raise Exception('The provided "frame" is not a valid frame.')

        is_mask = False
        if NumpyFrameHelper.is_alpha(frame):
            is_mask = True
            # We ensure it is a normalized alpha frame to store it
            frame = NumpyFrameHelper.as_alpha(frame)
        elif NumpyFrameHelper.is_rgb(frame):
            # We ensure it is a not normalized normal frame to store it
            frame = NumpyFrameHelper.as_rgb(frame)
        else:
            raise Exception('The provided "frame" is not an alpha nor a rgb frame.')

        self.frame = frame
        self._is_mask = is_mask

    def as_mask(self, masking_method: FrameMaskingMethod = FrameMaskingMethod.MEAN):
        """
        Return the frame as a mask by applying the 'masking_method'
        if necessary.
        """
        if self.is_mask:
            return self.frame
        
        return NumpyFrameHelper.as_alpha(self.frame, do_normalize = True, masking_method = masking_method)
    
    def inverted(self):
        """
        Return the frame inverted. This method does not modify
        the object itself.
        """
        return NumpyFrameHelper.invert(self.frame)
    
    def normalized(self):
        """
        Return the frame normalized (with values between 0.0
        and 1.0). This method does not modify the object
        itself.
        """
        return NumpyFrameHelper.normalize(self.frame)

    def denormalized(self):
        """
        Return the frame denormalized (with values between 0
        and 255). This method does not modify the object
        itself.
        """
        return NumpyFrameHelper.denormalize(self.frame)

